#ifndef FGGL_MATH_TYPES_H
#define FGGL_MATH_TYPES_H

#include <tuple>

#include <glm/ext/matrix_transform.hpp>
#include <glm/glm.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>

namespace fggl::math {

	// math types (aliased for ease of use)
	using vec4 = glm::vec4;
	using vec3 = glm::vec3;
	using vec2 = glm::vec2;
	using mat4 = glm::mat4;
	using quat = glm::quat;

    // fastFloor from OpenSimplex2
    inline int fastFloor(double x) {
        int xi = (int)x;
        return x < xi ? xi - 1 : xi;
    }

	// reference vectors
	constexpr vec3 UP { 0.0f, 1.0f, 0.0f };
	constexpr vec3 FORWARD { 1.0f, 0.0f, 0.0f };
	constexpr vec3 RIGHT { 0.0f, 0.0f, 1.0f };

	inline glm::mat4 modelMatrix( const vec3 offset, const quat rotation ) {
		return glm::translate( glm::mat4(1.0f), offset ) * glm::toMat4( rotation );
	}

	inline glm::mat4 modelMatrix( glm::vec3 offset, glm::vec3 eulerAngles ) {
		return modelMatrix( offset, glm::quat( eulerAngles ) );
	}

	struct Transform {
        constexpr static const char name[] = "Transform";

		Transform() : m_local(1.0f), m_origin(0.0f), m_model(1.0f), m_rotation() {
		}

		// local reference vectors
		[[nodiscard]]
		inline vec3 up() const {
			return vec4( UP, 1.0 ) * m_local;
		}

		[[nodiscard]]
		inline vec3 forward() const {
			return vec4( FORWARD, 1.0 ) * m_local;
		}

		[[nodiscard]]
		inline vec3 right() const {
			return vec4( RIGHT, 1.0 ) * m_local;
		}

		inline void translate(const vec3 change) {
			m_origin += change;
			update();
		}

		inline void origin(const vec3 pos) {
			m_origin = pos;
			update();
		}

		[[nodiscard]]
		inline vec3 origin() const {
			return m_origin;
		}

		inline void euler(vec3 angles) {
			m_rotation = quat( angles );
			update();
		}

		[[nodiscard]]
		inline glm::vec3 euler() const {
			return glm::eulerAngles(m_rotation);
		}

		inline mat4 model() const {
			mat4 tmp(1.0f);
			tmp = glm::translate(tmp, m_origin);
			tmp = tmp * glm::toMat4(m_rotation);

			return tmp;
		}

		private:
			mat4 m_local; // us -> parent
			mat4 m_model; // us -> world
			vec3 m_origin;
			quat m_rotation;

			inline void update() {
				mat4 t(1.0f);
				t *= glm::toMat4(m_rotation);
				t = glm::translate( t, m_origin );
				m_local = t;
			}
	};

}


// feels a bit strange to be doing this...
namespace glm {
	inline bool operator<(const vec3& lhs, const vec3& rhs) {
		return std::tie( lhs.x, lhs.y, lhs.z )
		      < std::tie( rhs.x, rhs.y, rhs.z );
	}
}


#endif
